/*
 * Pilot Datebook processing utility
 * Matthias Hessler <pilot-datebook@mhessler.de> December 2000
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */


#include "pilot-datebook-joblist.h"


/* Data structures */



/* Function definitions */

/* Public */
int main (int argc, char **argv);
void usage (char * progname, char * opt_arg);

